
{*******************************************************}
{                                                       }
{       JADScript compiler by Y [27-11-00]              }
{       Abstract textual code generator                 }
{                                                       }
{       Copyright (c) 1999-2000 YALLIE, Inc.            }
{       All Rights Reserved.                            }
{                                                       }
{*******************************************************}

unit CCodeGenerator; {$DEFINE DEBUG}

interface

uses Objects, CConstants;

const
  cBufferSize = 16384;

type
  TCodeStyle = (stC, stPascal);

  PCodeGenerator = ^TCodeGenerator;
  TCodeGenerator = object(TObject)
    FileName: String;
    Style: TCodeStyle;
    OutStream: TBufStream;
    Status: Integer;

    {public}
    constructor Init(aFileName: String; aStyle: TCodeStyle);
    destructor Done; virtual;

    procedure Emit(Line: String);
    procedure EmitLine(Line: String);
    procedure Emitln;

    procedure EmitComment(Line: String); virtual;
    procedure EmitStringConstant(Line: String; Quotes: SmallString); virtual;

    procedure EmitFirst; virtual;
    procedure EmitLast; virtual;
  end;

implementation

constructor TCodeGenerator.Init(aFileName: String; aStyle: TCodeStyle);
begin
  Style := aStyle;
  FileName := aFileName;
  OutStream.Init(FileName, stCreate, cBufferSize);
  if (OutStream.Status = 0) then
  begin
    EmitLn;
    EmitFirst;
  end;
end; {TCodeGenerator.Init}

destructor TCodeGenerator.Done;
begin
  EmitLast;

  OutStream.Done;
  Status := OutStream.Status;
end; {TCodeGenerator.Done}

procedure TCodeGenerator.EmitFirst;
begin
  EmitComment('TCodeGenerator/JADScript version ' +
    Version + ' by Y [27-11-00]');
  EmitComment('Do not edit manually.');
end; {TCodeGenerator.EmitFirst}

procedure TCodeGenerator.EmitLast;
begin
  EmitComment('End of file.');
end; {TCodeGenerator.EmitLast}

procedure TCodeGenerator.Emit(Line: String);
begin
  if (OutStream.Status = 0) then
    OutStream.Write(Line[1], Length(Line));

  Status := OutStream.Status;
end; {TCodeGenerator.Emit}

procedure TCodeGenerator.EmitLine(Line: String);
const
  Temp: String = #13#10;
begin
  if (OutStream.Status = 0) then
  begin
    {write line}
    OutStream.Write(Line[1], Length(Line));

    {write CR/LF}
    if (OutStream.Status = 0) then
      OutStream.Write(Temp[1], 2);
  end;
  Status := OutStream.Status;
end; {TCodeGenerator.EmitLine}

procedure TCodeGenerator.EmitLn;
begin
  EmitLine('');
end; {TCodeGenerator.EmitLn}

procedure TCodeGenerator.EmitComment(Line: String);
begin
  if Style = stC then
    EmitLine('// ' + Line)
  else
    EmitLine('{ ' + Line + ' }');
end; {TCodeGenerator.EmitComment}

procedure TCodeGenerator.EmitStringConstant(Line: String; Quotes: SmallString);
var
  Quote1, Quote2, A: Char;
  NewLine: String;
  i: Integer;
begin
  if Quotes <> '' then
  begin
    Emit(Quotes[1]);
    Quote1 := Quotes[1];
    Quote2 := Quotes[2];
  end
  else
  begin
    Quote1 := #0;
    Quote2 := #0;
  end;

  {fix slashes and quotes}
  NewLine := '';
  for i := 1 to Length(Line) do
  begin
    A := Line[i];
    if (A = Quote1) or (A = Quote2) or (A = '\') or (A < ' ') then
      NewLine := NewLine + '\';
    NewLine := NewLine + A;
  end;
  Emit(NewLine);

  if Quotes <> '' then
    Emit(Quotes[2]);
end; {TCodeGenerator.EmitComment}

end {CCodeGenerator}.
