
{*******************************************************}
{                                                       }
{       JADScript compiler by Y [27-11-00]              }
{       Pascal code generator                           }
{                                                       }
{       Copyright (c) 1999-2000 YALLIE, Inc.            }
{       All Rights Reserved.                            }
{                                                       }
{*******************************************************}

unit CCodePascal; {$DEFINE DEBUG}

interface

uses Objects, CConstants, CCodeGenerator, CLists;

const
  cTab = '  ';

type
  PPascalCodeGenerator = ^TPascalCodeGenerator;
  TPascalCodeGenerator = object(TCodeGenerator)
    constructor Init(aFileName: String);
    procedure EmitFirst; virtual;
    procedure EmitLast; virtual;
    procedure EmitBorlandComment(Line: String);

    {main}
    procedure Generate(JadObject: PJadObject);

    {utility subroutines}
    function MangleName(Name: String): String;
    {function UnmangleName(Name: String): String;}
    function GetMargin(Level: Integer): String;
    procedure Error(Message: String);
    procedure Warning(Message: String);

    {code generation}
    procedure EmitHeader(JadObject: PJadObject);
    procedure GenerateProcedure(JadObject: PJadObject);
    procedure GenerateBlock(JadObject: PJadObject; Level: Integer);
    procedure GenerateProcess(JadObject: PJadObject; Level: Integer);
    procedure GenerateVars(JadObject: PJadObject; Level: Integer);
    procedure GenerateComment(JadObject: PJadObject; Level: Integer);
    procedure GenerateCondition(JadObject: PJadObject; Level: Integer);
    procedure GenerateExit(JadObject: PJadObject; Level: Integer);
    procedure ProcessAllWaysFrom(JadObject: PJadObject; Level: Integer);
  end;

implementation

constructor TPascalCodeGenerator.Init(aFileName: String);
begin
  inherited Init(aFileName, stPascal);
end; {TPascalCodeGenerator.Init}

procedure TPascalCodeGenerator.EmitBorlandComment(Line: String);
var
  i: Integer;
  Temp: String;
begin
  Temp := '';
  if Line = '*' then
    for i := 1 to 55 do
      Temp := Temp + '*'
  else
  begin
    for i := 1 to 7 do
      Temp := Temp + ' ';
    Temp := Temp + Copy(Line, 1, 42);
    if Length(Line) > 45 then
      Temp := Temp + '...'
    else
      Temp := Temp + Copy(Line, 43, 3);
    for i := Length(Temp) to 54 do
      Temp := Temp + ' ';
  end;
  EmitLine('{' + Temp + '}');
end; {TPascalCodeGenerator.EmitBorlandComment}

procedure TPascalCodeGenerator.EmitFirst;
begin
  {nothing}
end; {TPascalCodeGenerator.EmitFirst}

procedure TPascalCodeGenerator.EmitLast;
begin
  {nothing}
end; {TPascalCodeGenerator.EmitLast}

procedure TPascalCodeGenerator.EmitHeader(JadObject: PJadObject);
begin
  EmitBorlandComment('*');
  EmitBorlandComment('');

  if JadObject <> nil then
  begin
    EmitBorlandComment('Project: ' + JadObject^.GetAttribute('Name'));
    EmitBorlandComment('Author: ' + JadObject^.GetAttribute('Author'));
    EmitBorlandComment('Date: ' + JadObject^.GetAttribute('Date'));
  end
  else
    EmitBorlandComment('Project: (null)');

  EmitBorlandComment('Code: TPascalCodeGenerator ' + Version);
  EmitBorlandComment('');
  EmitBorlandComment('Copyright (c) 1999-2000 YALLIE, Inc.');
  EmitBorlandComment('All Rights Reserved.');
  EmitBorlandComment('');
  EmitBorlandComment('*');
  EmitLn;
  EmitLine('{$Ifdef Win32} {$AppType Console}');
  EmitLn;
  EmitLine('uses Args;');
  EmitLn;
end; {TPascalCodeGenerator.EmitHeader}

{MangleName removes non-alphanumeric characters, UnmangleName undoes it
function TPascalCodeGenerator.MangleName(Name: String): String;
var
  Temp: String;
  i: Integer;
begin
  Temp := '';

  if Name <> '' then
  begin
    if Name[1] in ['A'..'Z', 'a'..'z'] then
      Temp := Temp + Name[1]
    else
      Temp := Temp + '_' + CharToHex(Name[1]);

    for i := 2 to Length(Name) do
      if Name[i] in ['A'..'Z', 'a'..'z', '0'..'9'] then
        Temp := Temp + Name[i]
      else
        Temp := Temp + '_' + CharToHex(Name[i]);
  end;

  MangleName := Temp;
end; {TPascalCodeGenerator.MangleName}

{function TPascalCodeGenerator.UnmangleName(Name: String): String;
var
  Temp: String;
  i: Integer;
begin
  Temp := '';

  while Name <> '' do
    if Name[1] <> '_' then
    begin
      Temp := Temp + Name[1];
      Delete(Name, 1, 1);
    end
    else
    begin
      Delete(Name, 1, 1);
      Temp := Temp + HexToChar(Name);
      Delete(Name, 1, 2);
    end;

  UnmangleName := Temp;
end; {TPascalCodeGenerator.UnmangleName}

function TPascalCodeGenerator.MangleName(Name: String): String;
var
  Temp: String;
  i: Integer;
begin
  Temp := '';

  if Name <> '' then
  begin
    if Name[1] in ['0'..'9'] then
      Temp := '_';

    for i := 1 to Length(Name) do
      if Name[i] in ['A'..'Z', 'a'..'z', '0'..'9'] then
        Temp := Temp + Name[i]
      else
        Temp := Temp + CharToAlpha(Name[i], en1251);
  end;

  MangleName := Temp;
end; {TPascalCodeGenerator.MangleName}

function TPascalCodeGenerator.GetMargin(Level: Integer): String;
var
  i: Integer;
  Temp: String;
begin
  Temp := '';
  for i := 1 to Level do
    Temp := Temp + cTab;
  GetMargin := Temp;
end; {TPascalCodeGenerator.GetMargin}

procedure TPascalCodeGenerator.Error(Message: String);
begin
  Writeln('Fatal: ' + Message);
  Halt(Byte(-1));
end; {TPascalCodeGenerator.Error}

procedure TPascalCodeGenerator.Warning(Message: String);
begin
  Writeln('Warning: ' + Message);
end; {TPascalCodeGenerator.Warning}

procedure TPascalCodeGenerator.Generate(JadObject: PJadObject);
var
  List: PJadObject;
  Name: String;
  i: Integer;
begin
  EmitHeader(JadObject);

  if JadObject <> nil then
  begin
    {emit forward declarations to ensure all is right}
    EmitComment('Forward declarations');
    for i := 1 to JadObject^.GetObjectsCount do
    begin
      List := JadObject^.ObjectAt(i - 1);
      if List <> nil then
        Name := MangleName(List^.GetAttribute('Name'))
      else
        Error('list number ' + IntToStr(i - 1) + ' not found');

      EmitLine('procedure ' + Name + '; forward;');
    end;
    EmitLn;

    {emit all procedures}
    EmitComment('Actual declarations');
    EmitLn;

    for i := 1 to JadObject^.GetObjectsCount do
    begin
      List := JadObject^.ObjectAt(i - 1);
      GenerateProcedure(List);
    end;

    {emit program body}
    EmitLine('begin');
    Emit(cTab);

    List := JadObject^.ObjectAt(0);
    if List <> nil then
      EmitLine(MangleName(List^.GetAttribute('Name')) + ';')
    else
      EmitComment('Empty project: no lists');

    EmitLine('{$Else Win32}'#13#10 +
      'begin'#13#10 +
      cTab + 'Writeln(''This program requires Delphi ' +
        '32-bit compiler DCC32.EXE.'');'#13#10 +
      '{$Endif Win32}');

    EmitLine('end {' + JadObject^.GetAttribute('Name') + '}.');
  end
  else
    EmitComment('Empty project: no lists');
end; {TPascalCodeGenerator.Generate}

procedure TPascalCodeGenerator.GenerateProcedure(JadObject: PJadObject);
var
  TempObj, Entry: PJadObject;
  i, Count, Tag: Integer;
  Name, TempStr: String;
  First: Boolean; {general purpose flag ;}
begin
  {comments}
  First := True;
  for i := 1 to JadObject^.GetObjectsCount do
  begin
    TempObj := JadObject^.ObjectAt(i - 1);
    if TempObj <> nil then
      if (LowerCase(TempObj^.GetType) = 'comment') then
      begin
        First := False;
        GenerateComment(TempObj, 1);
      end;
  end;
  if not First then
    Emitln;

  {declaration}
  Count := JadObject^.GetObjectsCount;
  Name := MangleName(JadObject^.GetAttribute('Name'));
  EmitLine('procedure ' + Name + ';');

  {labels: every block that has more than one entry should be labeled}
  First := True;
  for i := 1 to JadObject^.GetObjectsCount do
  begin
    TempObj := JadObject^.ObjectAt(i - 1);
    if TempObj <> nil then
      if (LowerCase(TempObj^.GetType) = 'block') or
        (LowerCase(TempObj^.GetType) = 'process') or
        (LowerCase(TempObj^.GetType) = 'condition') or
        (LowerCase(TempObj^.GetType) = 'exit') then
      begin
        if StrToInt(TempObj^.GetAttribute('Inputs')) > 1 then
        begin
          TempStr := 'Label' + TempObj^.GetAttribute('Tag');
          TempObj^.AddAttribute('__Label', TempStr);
          if First then
          begin
            EmitLine('label { ' + GetHmmMessage + ' }');
            Emit(cTab);
            First := False;
          end
          else
            Emit(', ');
          Emit(TempStr);
        end;
      end;
  end;
  if not First then
    EmitLine(';');

  {constants}
  First := True;
  for i := 1 to JadObject^.GetObjectsCount do
  begin
    TempObj := JadObject^.ObjectAt(i - 1);
    if TempObj <> nil then
      if (LowerCase(TempObj^.GetType) = 'constants') then
      begin
        if First then
          EmitLine('const');
        First := False;
        GenerateVars(TempObj, 1);
      end;
  end;

  {variables}
  First := True;
  for i := 1 to JadObject^.GetObjectsCount do
  begin
    TempObj := JadObject^.ObjectAt(i - 1);
    if TempObj <> nil then
      if (LowerCase(TempObj^.GetType) = 'variables') then
      begin
        if First then
          EmitLine('var');
        First := False;
        GenerateVars(TempObj, 1);
      end;
  end;

  {body}
  EmitLine('begin');

  {find the first entry point - all others are erroneous}
  Entry := JadObject^.FindObject('Entry', '', ''); {attribute does not mean}
  if Entry = nil then
    Error('no entry point to code in procedure ' + Name);

  {follow links from this object}
  ProcessAllWaysFrom(Entry, 1);
(*  {find the first block: only one output allowed for Entry object}
  Tag := StrToInt(Entry^.GetAttribute('Output0'));
  TempObj := JadObject^.FindObject('Connector', 'Tag', IntToStr(Tag));
  if TempObj = nil then
    Error('unassigned connector (Tag = ' + IntToStr(Tag) +
      ') in procedure ' + Name);

  Tag := StrToInt(TempObj^.GetAttribute('TagEnd')); {get block tag}
  TempObj := JadObject^.FindObject('Block', 'Tag', IntToStr(Tag)); {get block}
  if TempObj = nil then
    Error('unassigned block (Tag = ' + IntToStr(Tag) +
      ') in procedure ' + Name);

  GenerateBlock(TempObj, 1);*)

  {end of declaration}
  EmitLine('end; {' + MangleName(JadObject^.GetAttribute('Name')) + '}');
  EmitLn;
end; {TPascalCodeGenerator.GenerateProcedure}

procedure TPascalCodeGenerator.GenerateBlock(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i: Integer;
  Temp: String;
begin
  {check if already processed}
  if JadObject^.GetAttribute('__Processed') <> '' then
  begin
    {if this block is already processed, then it was labeled before}
    Temp := JadObject^.GetAttribute('__Label');
    EmitLine(GetMargin(Level) + 'goto ' + Temp + '; { ' +
      GetHmmMessage + ' }');
    Exit;
  end;

  {mark block as processed}
  JadObject^.AddAttribute('__Processed', '1');

  {emit label if defined}
  Margin := GetMargin(Level - 1);
  Temp := JadObject^.GetAttribute('__Label');
  if Temp <> '' then
    EmitLine(Margin + Temp + ': { ' + GetHmmMessage + ' }');

  {emit code}
  Margin := GetMargin(Level);
  Count := StrToInt(JadObject^.GetAttribute('Strings'));
  for i := 1 to Count do
  begin
    Temp := JadObject^.GetAttribute('String' + IntToStr(i - 1));
    EmitLine(Margin + Temp);
  end;

  {follow links}
  ProcessAllWaysFrom(JadObject, Level);
end; {TPascalCodeGenerator.GenerateBlock}

procedure TPascalCodeGenerator.GenerateProcess(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i: Integer;
  Temp: String;
begin
  {check if already processed}
  if JadObject^.GetAttribute('__Processed') <> '' then
  begin
    {if this process is already processed, then it was labeled before}
    Temp := JadObject^.GetAttribute('__Label');
    EmitLine(GetMargin(Level) + 'goto ' + Temp + '; { ' +
      GetHmmMessage + ' }');
    Exit;
  end;

  {mark block as processed}
  JadObject^.AddAttribute('__Processed', '1');

  {emit label if defined}
  Margin := GetMargin(Level - 1);
  Temp := JadObject^.GetAttribute('__Label');
  if Temp <> '' then
    EmitLine(Margin + Temp + ': { ' + GetHmmMessage + ' }');

  {emit code}
  Margin := GetMargin(Level);
  Count := StrToInt(JadObject^.GetAttribute('Strings'));
  if Count > 1 then
    Warning('process ' + JadObject^.GetAttribute('Name') + ' has too ' +
      'much lines, only the first is actual')
  else if Count = 0 then
    Warning('process ' + JadObject^.GetAttribute('Name') + ' is empty');

  Temp := JadObject^.GetAttribute('String0');
  if Temp <> '' then
    EmitLine(Margin + Temp + ';');

  {follow links}
  ProcessAllWaysFrom(JadObject, Level);
end; {TPascalCodeGenerator.GenerateProcess}

procedure TPascalCodeGenerator.GenerateVars(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i: Integer;
  Temp: String;
begin
  {emit code}
  Margin := GetMargin(Level);
  Count := StrToInt(JadObject^.GetAttribute('Strings'));
  for i := 1 to Count do
  begin
    Temp := JadObject^.GetAttribute('String' + IntToStr(i - 1));
    EmitLine(Margin + Temp);
  end;
end; {TPascalCodeGenerator.GenerateVars}

procedure TPascalCodeGenerator.GenerateComment(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i: Integer;
  Temp: String;
begin
  {emit code}
  Margin := GetMargin(Level);
  Count := StrToInt(JadObject^.GetAttribute('Strings'));
  if Count > 0 then
    Emit('{ ');

  for i := 1 to Count do
  begin
    Temp := JadObject^.GetAttribute('String' + IntToStr(i - 1));
    if i = 1 then
      Emit(Margin + Temp)
    else
    begin
      EmitLn;
      Emit(Margin + '  ' + Temp);
    end;
  end;

  if Count > 0 then
    EmitLine(' }');
end; {TPascalCodeGenerator.GenerateComment}

procedure TPascalCodeGenerator.GenerateCondition(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i, Tag: Integer;
  Temp: String;
begin
  {check if already processed}
  if JadObject^.GetAttribute('__Processed') <> '' then
  begin
    {if this block is already processed, then it was labeled before}
    Temp := JadObject^.GetAttribute('__Label');
    EmitLine(GetMargin(Level) + 'goto ' + Temp + '; { ' +
      GetHmmMessage + ' }');
    Exit;
  end;

  {mark condition as processed}
  JadObject^.AddAttribute('__Processed', '1');

  {emit label if defined}
  Margin := GetMargin(Level - 1);
  Temp := JadObject^.GetAttribute('__Label');
  if Temp <> '' then
    EmitLine(Margin + Temp + ': { ' + GetHmmMessage + ' }');

  {emit 'if (Expression) then'}
  Margin := GetMargin(Level);
  Count := StrToInt(JadObject^.GetAttribute('Strings'));
  Emit(Margin);

  if Count > 0 then
    Emit('if (')
  else
    EmitLine('{ this condition is always True }');

  for i := 1 to Count do
  begin
    Temp := JadObject^.GetAttribute('String' + IntToStr(i - 1));
    if i = 1 then
      Emit(Temp)
    else
    begin
      EmitLn;
      Emit(Margin + '  ' + Temp);
    end;
  end;

  if Count > 0 then
    EmitLine(') then');

  {emit block}
  EmitLine(Margin + 'begin');
  ProcessAllWaysFrom(JadObject, Level + 1);
  EmitLine(Margin + 'end;');
end; {TPascalCodeGenerator.GenerateCondition}

procedure TPascalCodeGenerator.GenerateExit(JadObject: PJadObject; Level: Integer);
var
  Margin: String;
  Count, i: Integer;
  Temp: String;
begin
  {check if already processed}
  if JadObject^.GetAttribute('__Processed') <> '' then
  begin
    {if this exit is already processed, then it was labeled before}
    Temp := JadObject^.GetAttribute('__Label');
    EmitLine(GetMargin(Level) + 'goto ' + Temp + '; { ' +
      GetHmmMessage + ' }');
    Exit;
  end;

  {mark exit as processed}
  JadObject^.AddAttribute('__Processed', '1');

  {emit label if defined}
  Margin := GetMargin(Level - 1);
  Temp := JadObject^.GetAttribute('__Label');
  if Temp <> '' then
    EmitLine(Margin + Temp + ': { ' + GetHmmMessage + ' }');

  {emit code}
  Margin := GetMargin(Level);
  EmitLine(Margin + 'Exit; { ' + JadObject^.GetAttribute('Name') + ' }');
end; {TPascalCodeGenerator.GenerateExit}

procedure TPascalCodeGenerator.ProcessAllWaysFrom(JadObject: PJadObject; Level: Integer);
var
  Margin, Name, ObjType: String;
  Count, i, Tag: Integer;
  TempObj, Page, NextBlock: PJadObject;
begin
  NextBlock := nil;
  Name := LowerCase(JadObject^.GetType + ' (Tag = ' +
    JadObject^.GetAttribute('Tag') + ', Name = "' +
    JadObject^.GetAttribute('Name') + '")');

  Page := JadObject^.GetParent;
  if Page = nil then
    Error('internal error: no parent for ' + Name);

  Count := StrToInt(JadObject^.GetAttribute('Outputs'));
  if (Count = 0) then
    Error('no ways from ' + Name);

  {explore links that lead to 'Condition's first}
  for i := 1 to Count do
  begin
    {get connector}
    TempObj := Page^.FindObject('Connector', 'Tag',
      JadObject^.GetAttribute('Output' + IntToStr(i - 1)));
    if TempObj = nil then
      Error('connector expected in ' + Name);

    {get the next object}
    TempObj := Page^.FindObject('', 'Tag',
      TempObj^.GetAttribute('TagEnd'));
    if TempObj = nil then
      Error('connector leads to nowhere in ' + Name);
    if LowerCase(TempObj^.GetType) = 'condition' then
      GenerateCondition(TempObj, Level)
    else
      if NextBlock = nil then
        NextBlock := TempObj
      else
        Error('two ways lead from ' + Name);
  end;

  {generate last block}
  if NextBlock <> nil then
  begin
    ObjType := LowerCase(NextBlock^.GetType);
    if ObjType = 'block' then
      GenerateBlock(NextBlock, Level)
    else if ObjType = 'process' then
      GenerateProcess(NextBlock, Level)
    else if ObjType = 'exit' then
      GenerateExit(NextBlock, Level)
    else
      Error('element of type "' + ObjType + '" is not allowed in ' + Name);
  end
  else
    Warning('no unconditional way from ' + Name);
end; {TPascalCodeGenerator.ProcessAllWays}

end {CCodePascal}.
