
{*******************************************************}
{                                                       }
{       JADScript compiler by Y [27-11-00]              }
{       List classes (string list, object list)         }
{                                                       }
{       Copyright (c) 1999-2000 YALLIE, Inc.            }
{       All Rights Reserved.                            }
{                                                       }
{*******************************************************}

unit CLists;

{$I CDEFINES.PAS} {$IFDEF TPC} {$N+} {$ENDIF}

interface

uses Objects, CConstants;

type

  PUnsortedCollection = ^TUnsortedCollection;
  TUnsortedCollection = object(TCollection)
    Duplicates: Boolean;
    constructor Init(ALimit, ADelta: Integer);
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    function KeyOf(Item: Pointer): Pointer; virtual;
    function Search(Key: Pointer; var Index: Integer): Boolean; virtual;
    procedure Insert(Item: Pointer); virtual;
  end;

  PUnsortedStringList = ^TUnsortedStringList;
  TUnsortedStringList = object(TUnsortedCollection)
    CaseSensitive: Boolean;
    constructor Init(ALimit, ADelta: Integer);
    constructor Load(var S: TStream);
    procedure Store(var S: TStream);
    function Add(S: String): Integer;
    function AddList(L: PUnsortedStringList): Integer;
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    procedure FreeItem(Item: Pointer); virtual;
    function StringAt(Index: Integer): String;
    procedure Clear;
  end;

  PNameValue = ^TNameValue;
  TNameValue = object(TObject)
    Name, Value: PString;

    constructor Init(aName, aValue: String);
    destructor Done; virtual;
    function GetName: String;
    function GetValue: String;
  end;

  PNameValueList = ^TNameValueList;
  TNameValueList = object(TUnsortedCollection)
    {internal constant}
    fCompareValue: Boolean;

    NameCaseSensitive, ValueCaseSensitive: Boolean;
    constructor Init(ALimit, ADelta: Integer);
    function Add(Name, Value: String): Integer;
    procedure FreeItem(Item: Pointer); virtual;
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    function NameAt(Index: Integer): String;
    function ValueAt(Index: Integer): String;
    function SearchName(Name: String; var Index: Integer): Boolean;
    function GetValueOf(Name: String; var Value: String): Boolean;
  end;

  PJadObjectList = ^TJadObjectList;

  PJadObject = ^TJadObject;
  TJadObject = object(TObject)
    fType: PString;
    fAttributes: PNameValueList;
    fObjects: PJadObjectList;
    fParent: PJadObject;
    CaseSensitive: Boolean;
    constructor Init;
    destructor Done; virtual;

    procedure SetType(aType: String);
    function GetType: String;
    function GetParent: PJadObject;

    function AddAttribute(Name, Value: String): Integer;
    function GetAttributesCount: Integer;
    function NameAt(Index: Integer): String;
    function ValueAt(Index: Integer): String;
    function GetAttribute(Name: String): String;

    function AddObject(JadObject: PJadObject): Integer;
    function ObjectAt(Index: Integer): PJadObject;
    function GetObjectsCount: Integer;

    function FindObject(ObjType, AttrName, AttrValue: String): PJadObject;
  end;

  TJadObjectList = object(TUnsortedCollection)
    function Add(JadObject: PJadObject): Integer;
    function ObjectAt(Index: Integer): PJadObject;
  end;

procedure RegisterLists;

implementation

{TStream records}

const
  RUnsortedStringList: TStreamRec = (
     ObjType: 101;
     VmtLink: Ofs(TypeOf(TUnsortedStringList)^);
     Load:    @TUnsortedStringList.Load;
     Store:   @TUnsortedStringList.Store
  );

{TUnsortedCollection}

constructor TUnsortedCollection.Init(ALimit, ADelta: Integer);
begin
  inherited Init(ALimit, ADelta);
  Duplicates := False;
end; {TUnsortedCollection.Init}

function TUnsortedCollection.Compare(Key1, Key2: Pointer): Integer;
begin
  Abstract;
  {disable FPC warnings}
  Compare := Integer(Key1 = Key2);
end; {TUnsortedCollection.Compare}

procedure TUnsortedCollection.Insert(Item: Pointer);
var
  I: Integer;
begin
  if not Search(KeyOf(Item), I) then
    AtInsert(I, Item)
  else if Duplicates then
    AtInsert(Count, Item);
end; {TUnsortedCollection.Insert}

function TUnsortedCollection.KeyOf(Item: Pointer): Pointer;
begin
  KeyOf := Item;
end; {TUnsortedCollection.KeyOf}

function TUnsortedCollection.Search(Key: Pointer; var Index: Integer): Boolean;
var
  i: Integer;
  Temp: Boolean;
begin
  Temp := False;
  i := 0;
  while (not Temp) and (i < Count) do
  begin
    if (Compare(KeyOf(Items^[i]), Key) = 0) then
    begin
      Temp := True;
      Index := i;
    end;
    Inc(i);
  end;
  if not Temp then
    Index := i;
  Search := Temp;
end; {TUnsortedCollection.Search}

{TUnsortedStringList}

constructor TUnsortedStringList.Init(ALimit, ADelta: Integer);
begin
  inherited Init(ALimit, ADelta);
  CaseSensitive := False;
end; {TUnsortedStringList.Init}

constructor TUnsortedStringList.Load(var S: TStream);
var
  aCount, i: Integer;
  TempString: String;
begin
  inherited Init(64, 16);
  S.Read(aCount, SizeOf(aCount));

  for i := 1 to aCount do
  begin
    S.Read(TempString[0], 1);
    S.Read(TempString[1], Byte(TempString[0]));
    Add(TempString);
  end;
end; {TUnsortedStringList.Load}

procedure TUnsortedStringList.Store(var S: TStream);
var
  aCount, i: Integer;
  TempString: String;
begin
  aCount := Self.Count;

  S.Write(aCount, SizeOf(aCount));
  for i := 1 to aCount do
  begin
    TempString := StringAt(i - 1);
    S.Write(TempString[0], Length(TempString) + 1);
  end;
end; {TUnsortedStringList.Store}

function TUnsortedStringList.Add(S: String): Integer;
var
  i: Integer;
  Item: Pointer;
begin
  Item := NewStr(S);
  if Duplicates then
  begin
    i := Count;
    AtInsert(Count, Item);
  end
  else
    if not Search(KeyOf(Item), i) then
      AtInsert(i, Item)
    else
      DisposeStr(Item);
  Add := i;
end; {TUnsortedStringList.Add}

function TUnsortedStringList.AddList(L: PUnsortedStringList): Integer;
var
  i, Temp: Integer;
begin
  if L <> nil then
    for i := 1 to L^.Count do
      Temp := Add(L^.StringAt(i - 1));

  AddList := Temp;
end; {TUnsortedStringList.AddList}

function CompareStringPointers(Key1, Key2: Pointer; CaseSensitive: Boolean):
  Integer; {Don't try to compare Nil pointers}
var
  i, j: Integer;
  P1, P2: PString;
  C1, C2: Char;
  L1, L2: Integer;
begin
  P1 := PString(Key1);
  P2 := PString(Key2);

  if (Length(P1^) < Length(P2^)) then
    J := Length(P1^)
  else
    J := Length(P2^);

  I := 1;

  if CaseSensitive then
  begin
    while (I < J) and (P1^[I] = P2^[I]) do
      Inc(I);
    C1 := P1^[I];
    C2 := P2^[I];
  end
  else
  begin
    while (I < J) and (LoCase(P1^[I]) = LoCase(P2^[I])) do
      Inc(I);
    C1 := LoCase(P1^[I]);
    C2 := LoCase(P2^[I]);
  end;

  if (I = J) then
  begin
    L1 := Length(P1^);
    L2 := Length(P2^);

    if (C1 < C2) then
      CompareStringPointers := -1
    else if (C1 > C2) then
      CompareStringPointers := 1
    else if (L1 > L2) then
      CompareStringPointers := 1
    else If (L1 < L2) then
      CompareStringPointers := -1
    else
      CompareStringPointers := 0;
  end
  else
    if (C1 < C2) then
      CompareStringPointers := -1
    else
      CompareStringPointers := 1;
end;

function TUnsortedStringList.Compare(Key1, Key2: Pointer): Integer;
var
  Temp: Integer;
begin
  if Key1 = nil then
    if Key2 = nil then
      Temp := 0
    else
      Temp := -1
  else
    if Key2 = nil then
      Temp := 1
    else
      Temp := CompareStringPointers(Key1, Key2, CaseSensitive);

  Compare := Temp;
end; {TUnsortedStringList.Compare}

procedure TUnsortedStringList.FreeItem(Item: Pointer);
begin
  DisposeStr(Item);
end; {TUnsortedStringList.FreeItem}

function TUnsortedStringList.StringAt(Index: Integer): String;
begin
  StringAt := PString(At(Index))^;
end; {TUnsortedStringList.StringAt}

procedure TUnsortedStringList.Clear;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    FreeItem(At(i));
  Count := 0;
end; {TUnsortedStringList.Clear}

{ TNameValue }

constructor TNameValue.Init(aName, aValue: String);
begin
  inherited Init;
  Name := NewStr(aName);
  Value := NewStr(aValue);
end; {TNameValue.Init}

destructor TNameValue.Done;
begin
  DisposeStr(Value);
  DisposeStr(Name);
  inherited Done;
end; {TNameValue.Done}

function TNameValue.GetName: String;
begin
  if Name <> nil then
    GetName := Name^
  else
    GetName := '';
end; {TNameValue.GetName}

function TNameValue.GetValue: String;
begin
  if Value <> nil then
    GetValue := Value^
  else
    GetValue := '';
end; {TNameValue.GetValue}

{ TNameValueList }

constructor TNameValueList.Init(aLimit, aDelta: Integer);
begin
  inherited Init(aLimit, aDelta);
  NameCaseSensitive := False;
  ValueCaseSensitive := True;
  fCompareValue := True;
end; {TNameValueList.Init}

function TNameValueList.Add(Name, Value: String): Integer;
var
  i: Integer;
  Item: Pointer;
begin
  Item := New(PNameValue, Init(Name, Value));
  AtInsert(Count, Item);
  Add := Count;
end; {TNameValueList.Add}

procedure TNameValueList.FreeItem(Item: Pointer);
var
  NameValue: PNameValue;
begin
  NameValue := Item;
  Dispose(NameValue, Done);
end; {TNameValueList.FreeItem}

function TNameValueList.Compare(Key1, Key2: Pointer): Integer;
var
  Name1, Name2, Value1, Value2: PString;
  Temp: Integer;
begin
  Name1 := PNameValue(Key1)^.Name;
  Name2 := PNameValue(Key2)^.Name;
  Value1 := PNameValue(Key1)^.Value;
  Value2 := PNameValue(Key2)^.Value;

  {compare names first}
  if Name1 = nil then
    if Name2 = nil then
      Temp := 0
    else
      Temp := -1
  else
    if Name2 = nil then
      Temp := 1
    else
      Temp := CompareStringPointers(Name1, Name2, NameCaseSensitive);

  {compare values only if names are equal}
  if (Temp = 0) and fCompareValue then
    if Value1 = nil then
      if Value2 = nil then
        Temp := 0
      else
        Temp := -1
    else
      if Value2 = nil then
        Temp := 1
      else
        Temp := CompareStringPointers(Value1, Value2, ValueCaseSensitive);

  Compare := Temp;
end; {TNameValueList.Compare}

function TNameValueList.NameAt(Index: Integer): String;
begin
  NameAt := PNameValue(At(Index))^.GetName;
end; {TNameValueList.NameAt}

function TNameValueList.ValueAt(Index: Integer): String;
begin
  ValueAt := PNameValue(At(Index))^.GetValue;
end; {TNameValueList.ValueAt}

function TNameValueList.SearchName(Name: String; var Index: Integer): Boolean;
var
  Temp: Boolean;
  NameValue: PNameValue;
begin
  NameValue := New(PNameValue, Init(Name, ''));
  {disable value to search by name only}
  fCompareValue := False;
  Temp := Search(NameValue, Index);
  {enable values again}
  fCompareValue := True;
  SearchName := Temp;
end; {TNameValueList.SearchName}

function TNameValueList.GetValueOf(Name: String; var Value: String): Boolean;
var
  Index: Integer;
  Temp: Boolean;
begin
  Temp := False;
  if SearchName(Name, Index) then
  begin
    Value := ValueAt(Index);
    Temp := True;
  end;

  GetValueOf := Temp;
end; {TNameValueList.GetValueOf}

{ TJadObject }

constructor TJadObject.Init;
begin
  fType := NewStr('');
  fAttributes := New(PNameValueList, Init(16, 8));
  fObjects := New(PJadObjectList, Init(32, 16));
  fParent := nil;
  CaseSensitive := False;
end; {TJadObject.Init}

destructor TJadObject.Done;
begin
  DisposeStr(fType);
  Dispose(fAttributes, Done);
  Dispose(fObjects, Done);
end; {TJadObject.Done}

procedure TJadObject.SetType(aType: String);
begin
  fType := NewStr(aType);
end; {TJadObject.SetType}

function TJadObject.GetType: String;
begin
  if fType <> nil then
    GetType := fType^
  else
    GetType := '';
end; {TJadObject.GetType}

function TJadObject.AddAttribute(Name, Value: String): Integer;
begin
  fAttributes^.Add(Name, Value);
end; {TJadObject.AddAttribute}

function TJadObject.GetAttributesCount: Integer;
begin
  GetAttributesCount := fAttributes^.Count;
end; {TJadObject.GetAttributesCount}

function TJadObject.NameAt(Index: Integer): String;
begin
  NameAt := fAttributes^.NameAt(Index);
end; {TJadObject.NameAt}

function TJadObject.ValueAt(Index: Integer): String;
begin
  ValueAt := fAttributes^.ValueAt(Index);
end; {TJadObject.ValueAt}

function TJadObject.GetAttribute(Name: String): String;
var
  Temp: String;
begin
  if fAttributes^.GetValueOf(Name, Temp) then
    GetAttribute := Temp
  else
    GetAttribute := '';
end; {TJadObject.GetAttribute}

function TJadObject.AddObject(JadObject: PJadObject): Integer;
begin
  JadObject^.fParent := @Self;
  AddObject := fObjects^.Add(JadObject);
end; {TJadObject.AddObject}

function TJadObject.ObjectAt(Index: Integer): PJadObject;
begin
  ObjectAt := fObjects^.ObjectAt(Index);
end; {TJadObject.ObjectAt}

function TJadObject.GetObjectsCount: Integer;
begin
  GetObjectsCount := fObjects^.Count;
end; {TJadObject.GetObjectsCount}

function TJadObject.FindObject(ObjType, AttrName, AttrValue: String): PJadObject;
var
  TempObj, Found: PJadObject;
  Count, i: Integer;

  function MyLowerCase(S: String): String;
  begin
    if CaseSensitive then
      MyLowerCase := S
    else
      MyLowerCase := LowerCase(S);
  end; {MyLowerCase}

begin
  Count := fObjects^.Count;
  Found := nil;

  {test all objects}
  for i := 1 to Count do
  begin
    TempObj := fObjects^.ObjectAt(i - 1);

    {if type is either right or not meaningful}
    if (ObjType = '') or (MyLowerCase(TempObj^.GetType) =
      MyLowerCase(ObjType)) then
      {if attribute is either right or not meaningful}
      if (AttrName = '') or (MyLowerCase(TempObj^.GetAttribute(AttrName)) =
        MyLowerCase(AttrValue)) then
      begin
        Found := TempObj;
        Break;
      end;
  end;

  FindObject := Found;
end; {TJadObject.FindObject}

function TJadObject.GetParent: PJadObject;
begin
  GetParent := fParent;
end; {TJadObject.GetParent}

{ TJadObjectList }

function TJadObjectList.Add(JadObject: PJadObject): Integer;
begin
  AtInsert(Count, JadObject);
  Add := Count;
end; {TJadObjectList.Add}

function TJadObjectList.ObjectAt(Index: Integer): PJadObject;
begin
  ObjectAt := At(Index);
end; {TJadObjectList.ObjectAt}

procedure RegisterLists;
begin
  RegisterType(RUnsortedStringList);
end; {RegisterLists}

begin
  RegisterLists;
end.
