
{*******************************************************}
{                                                       }
{       Pro2jds convertor by Y [28-11-00]               }
{       PRO file loader                                 }
{                                                       }
{       Copyright (c) 1999-2000 YALLIE, Inc.            }
{       All Rights Reserved.                            }
{                                                       }
{*******************************************************}

unit CProLoader;

{$I CDEFINES.PAS} {$IFDEF TPC} {$N+} {$ENDIF}

interface

uses Objects, CConstants, CLists;

type
  PProLoader = ^TProLoader;
  TProLoader = object(TBufStream)
    JadObject: PJadObject;
    LastItem: String;
    LastOffset, LastItemOffset: LongInt;
    constructor Init(FileName: FNameStr);
    destructor Done; virtual;
    procedure Fatal;

    procedure Load;
    procedure LoadDirEntry(Page: PJadObject);
    procedure LoadPage(Page: PJadObject);
    function LoadElement(Page: PJadObject): Boolean;
    procedure LoadBlock(Page: PJadObject; ElType: Byte);
    procedure LoadEntry(Page: PJadObject);
    procedure LoadVars(Page: PJadObject; ElType: Byte);
    procedure LoadComment(Page: PJadObject);
    procedure LoadConnector(Page: PJadObject; ElType: Byte);

    procedure AddString(fObject: PJadObject; Name: String);
    procedure AddChar(fObject: PJadObject; Name: String);
    procedure AddByte(fObject: PJadObject; Name: String);
    procedure AddWord(fObject: PJadObject; Name: String);
    procedure AddInteger(fObject: PJadObject; Name: String);
    procedure AddLongInt(fObject: PJadObject; Name: String);
    procedure AddDateTime(fObject: PJadObject; Name: String);
    procedure CheckSignature(Signature: String);
  end;

implementation

constructor TProLoader.Init(FileName: FNameStr);
begin
  inherited Init(FileName, stOpenRead, ReadBufferSize);
  JadObject := New(PJadObject, Init);
  LastItem := 'header';
  LastOffset := 0;
end; {TProLoader.Init}

destructor TProLoader.Done;
begin
  Dispose(JadObject, Done);
  inherited Done;
end; {TProLoader.Done}

procedure TProLoader.Fatal;
begin
  Writeln('Fatal: wrong file format at offset 0x' + IntToHex(LastOffset) +
    ' (' + LastItem + ', starts at 0x' + IntToHex(LastItemOffset)
    + ')');
  Halt(Byte(-1));
end; {TProLoader.Fatal}

procedure TProLoader.AddString(fObject: PJadObject; Name: String);
var
  Temp: String;
begin
  LastOffset := GetPos;
  Read(Temp, 1);
  Read(Temp[1], Byte(Temp[0]));
  fObject^.AddAttribute(Name, Temp);
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddString}

procedure TProLoader.AddChar(fObject: PJadObject; Name: String);
var
  Temp: String;
begin
  LastOffset := GetPos;
  Read(Temp[1], SizeOf(Char));
  Temp[0] := #1;
  fObject^.AddAttribute(Name, Temp);
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddChar}

procedure TProLoader.AddByte(fObject: PJadObject; Name: String);
var
  Temp: Byte;
begin
  LastOffset := GetPos;
  Read(Temp, SizeOf(Temp));
  fObject^.AddAttribute(Name, IntToStr(Temp));
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddByte}

procedure TProLoader.AddWord(fObject: PJadObject; Name: String);
var
  Temp: Word;
begin
  LastOffset := GetPos;
  Read(Temp, SizeOf(Temp));
  fObject^.AddAttribute(Name, IntToStr(Temp));
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddWord}

procedure TProLoader.AddInteger(fObject: PJadObject; Name: String);
var
  Temp: Integer;
begin
  LastOffset := GetPos;
  Read(Temp, SizeOf(Temp));
  fObject^.AddAttribute(Name, IntToStr(Temp));
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddInteger}

procedure TProLoader.AddLongInt(fObject: PJadObject; Name: String);
var
  Temp: LongInt;
begin
  LastOffset := GetPos;
  Read(Temp, SizeOf(Temp));
  fObject^.AddAttribute(Name, IntToStr(Temp));
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddLongInt}

procedure TProLoader.AddDateTime(fObject: PJadObject; Name: String);
var
  Temp: Double;
begin
  LastOffset := GetPos;
  Read(Temp, SizeOf(Temp));

  {we are too busy to convert anything ;}
  fObject^.AddAttribute(Name, DateTimeString(dtDateTime));
  if Status <> 0 then
    Fatal;
end; {TProLoader.AddDateTime}

procedure TProLoader.CheckSignature(Signature: String);
var
  Temp: String;
  Length: Char;
begin
  LastOffset := GetPos;
  Length := Signature[0];
  if Length > #0 then
  begin
    Temp[0] := Length;
    Read(Temp[1], Byte(Length));
    if Status <> 0 then
      Fatal;
    if Temp <> Signature then
      Fatal;
  end;
end; {TProLoader.CheckSignature}

procedure TProLoader.Load;
var
  Pages, i: Integer;
  Page: PJadObject;
begin
  LastItemOffset := GetPos;

  JadObject^.SetType(sIptCadType);
  JadObject^.AddAttribute('Version', '1.0');

  AddString(JadObject, 'Name');      {project name}
  AddDateTime(JadObject, 'Date');    {date/time - fake ;}
  AddString(JadObject, 'File');      {project file name}
  AddWord(JadObject, 'Pages');       {# of pages}
  AddString(JadObject, 'Author');    {author name}
  AddString(JadObject, 'Code');      {document code}

  {load page directory entries}
  Pages := StrToInt(JadObject^.GetAttribute('Pages'));
  for i := 1 to Pages do
  begin
    Page := New(PJadObject, Init);
    Page^.SetType('Page');
    LoadDirEntry(Page);
    JadObject^.AddObject(Page);
  end;

  CheckSignature(cStartPageArray);
  for i := 1 to Pages do
    LoadPage(JadObject^.ObjectAt(i - 1));

  CheckSignature(cEndPageArray);
end; {TProLoader.Load}

procedure TProLoader.LoadDirEntry(Page: PJadObject);
var
  Temp: String;
begin
  LastItem := 'page directory entry';
  LastItemOffset := GetPos;

  {page name is saved as String[50]}
  Read(Temp, cPageNameLength);
  Page^.AddAttribute('Name', Temp);
  if Status <> 0 then
    Fatal;

  {skip int64}
  Read(Temp, 8);
  AddByte(Page, 'Visible');
end; {TProLoader.LoadDirEntry}

procedure TProLoader.LoadPage(Page: PJadObject);
begin
  LastItemOffset := GetPos;
  LastItem := 'page';

  CheckSignature(cStartPage);
  AddWord(Page, 'Position');
  while LoadElement(Page) do;
end; {TProLoader.LoadPage}

function TProLoader.LoadElement(Page: PJadObject): Boolean;
var
  ElType: Byte;
  Result: Boolean;
begin
  LastItem := 'element';
  LastItemOffset := GetPos;

  Result := True;
  Read(ElType, SizeOf(ElType));

  {load specific elements}
  case ElType of
    Byte(cEndPage):
      Result := False; {end of page signature}
    elBlock, elProcess, elCondition:
      LoadBlock(Page, ElType);
    elEntryExit:
      LoadEntry(Page);
    elVariables, elConstants:
      LoadVars(Page, ElType);
    elComment:
      LoadComment(Page);
    elConnector:
      LoadConnector(Page, ElType);
    else
      Fatal;
  end;

  LoadElement := Result;
end; {TProLoader.LoadElement}

procedure TProLoader.LoadBlock(Page: PJadObject; ElType: Byte);
var
  Block: PJadObject;
  Count, i: LongInt;
begin
  LastItem := 'block';

  Block := New(PJadObject, Init);
  case ElType of
    elBlock: Block^.SetType('Block');
    elProcess: Block^.SetType('Process');
    elCondition: Block^.SetType('Condition');
  end;

  {Warning! This is undocumented: Condition doesn't have Name attribute}
  if ElType <> elCondition then
    AddString(Block, 'Name');

  AddLongint(Block, 'Tag');
  AddWord(Block, 'Top');
  AddWord(Block, 'Left');
  AddLongint(Block, 'Height');
  AddLongint(Block, 'Width');

  AddLongint(Block, 'Strings');
  Count := StrToInt(Block^.GetAttribute('Strings'));
  for i := 1 to Count do
    AddString(Block, 'String' + IntToStr(i - 1));

  AddLongInt(Block, 'Inputs');
  Count := StrToInt(Block^.GetAttribute('Inputs'));
  for i := 1 to Count do
    AddLongInt(Block, 'Input' + IntToStr(i - 1));
  AddLongInt(Block, 'Outputs');
  Count := StrToInt(Block^.GetAttribute('Outputs'));
  for i := 1 to Count do
    AddLongInt(Block, 'Output' + IntToStr(i - 1));

  Page^.AddObject(Block);
end; {TProLoader.LoadBlock}

procedure TProLoader.LoadEntry(Page: PJadObject);
var
  Entry: PJadObject;
  ElType: Byte;
  Count, i: LongInt;
begin
  LastItem := 'entry';

  Entry := New(PJadObject, Init);

  AddLongint(Entry, 'Tag');
  AddWord(Entry, 'Top');
  AddWord(Entry, 'Left');

  Read(ElType, SizeOf(ElType));
  case ElType of
    elEntry:
      Entry^.SetType('Entry');
    elExit:
      Entry^.SetType('Exit');
    else
      Fatal;
  end;

  AddChar(Entry, 'Name');
  AddLongInt(Entry, 'Inputs');
  Count := StrToInt(Entry^.GetAttribute('Inputs'));
  for i := 1 to Count do
    AddLongInt(Entry, 'Input' + IntToStr(i - 1));
  AddLongInt(Entry, 'Outputs');
  Count := StrToInt(Entry^.GetAttribute('Outputs'));
  for i := 1 to Count do
    AddLongInt(Entry, 'Output' + IntToStr(i - 1));

  Page^.AddObject(Entry);
end; {TProLoader.LoadEntry}

procedure TProLoader.LoadVars(Page: PJadObject; ElType: Byte);
var
  Vars: PJadObject;
  Count, i: LongInt;
begin
  LastItem := 'variables/constants';

  Vars := New(PJadObject, Init);

  if ElType = elConstants then
    Vars^.SetType('Constants')
  else
    Vars^.SetType('Variables');

  AddString(Vars, 'Name');
  AddLongInt(Vars, 'Tag');
  AddWord(Vars, 'Top');
  AddWord(Vars, 'Left');
  AddLongInt(Vars, 'Height');
  AddLongInt(Vars, 'Width');
  AddLongInt(Vars, 'Strings');

  Count := StrToInt(Vars^.GetAttribute('Strings'));
  for i := 1 to Count do
    AddString(Vars, 'String' + IntToStr(i - 1));

  Page^.AddObject(Vars);
end; {TProLoader.LoadVars}

procedure TProLoader.LoadComment(Page: PJadObject);
var
  Comment: PJadObject;
  Count, i: LongInt;
begin
  LastItem := 'comment';

  Comment := New(PJadObject, Init);
  Comment^.SetType('Comment');

  AddLongInt(Comment, 'Tag');
  AddWord(Comment, 'Top');
  AddWord(Comment, 'Left');
  AddLongInt(Comment, 'Height');
  AddLongInt(Comment, 'Width');
  AddLongInt(Comment, 'ArrowX');
  AddLongInt(Comment, 'ArrowY');
  AddLongInt(Comment, 'Strings');

  Count := StrToInt(Comment^.GetAttribute('Strings'));
  for i := 1 to Count do
    AddString(Comment, 'String' + IntToStr(i - 1));

  Page^.AddObject(Comment);
end; {TProLoader.LoadComment}

procedure TProLoader.LoadConnector(Page: PJadObject; ElType: Byte);
var
  Arrow: PJadObject;
  Count, i, Temp: LongInt;
begin
  LastItem := 'connector';

  Arrow := New(PJadObject, Init);
  Arrow^.SetType('Connector');

  AddLongInt(Arrow, 'Tag');
  AddLongInt(Arrow, 'TopLeft');
  AddLongInt(Arrow, 'TagStart');
  AddLongInt(Arrow, 'TagEnd');
  AddLongInt(Arrow, 'TopStart');
  AddLongInt(Arrow, 'TopEnd');

  Count := 0;
  Temp := 0;
  Read(Temp, SizeOf(Byte));

  while Temp = 0 do
  begin
    Read(Temp, SizeOf(Temp) - SizeOf(Byte));
    AddLongInt(Arrow, 'Left' + IntToStr(Count));
    AddLongInt(Arrow, 'Top' + IntToStr(Count));
    AddLongInt(Arrow, 'Right' + IntToStr(Count));
    AddLongInt(Arrow, 'Bottom' + IntToStr(Count));
    AddChar(Arrow, 'Type' + IntToStr(Count));
    Read(Temp, SizeOf(Byte));
    Inc(Count);
  end;
  Seek(GetPos - SizeOf(Byte));
  Arrow^.AddAttribute('Knees', IntToStr(Count));

  Page^.AddObject(Arrow);
end; {TProLoader.LoadConnector}

end {CProLoader}.
