
{*******************************************************}
{                                                       }
{       JADScript compiler by Y [27-11-00]              }
{       Just Another Language syntax analysis engine    }
{                                                       }
{       Copyright (c) 1999-2000 YALLIE, Inc.            }
{       All Rights Reserved.                            }
{                                                       }
{*******************************************************}

unit CSyntax; {$V-} {$DEFINE DEBUG}

interface

uses Dos, Objects, CConstants, CTokenStream, CLists;

type

  { TCompiler, syntax analysis tool }

  PJadsCompiler = ^TJadsCompiler;
  TJadsCompiler = object(TObject)
    Tokens: PTokenStream;
    JadObject: PJadObject;
    OldToken, Token: TToken;
    Lines: Integer;

    constructor Init(FileName: String);
    destructor Done; virtual;
    procedure Error(CompilerError: TCompilerError);

    procedure Compile;

    {invisible}
    function CompileObjectBody: Boolean;
    function CompileType: Boolean;
    function CompileAttributes: Boolean;
    function CompileObject: Boolean;
    function CompileAttribute: Boolean;
    function CompileEndOfFile: Boolean;
  end;

implementation

constructor TJadsCompiler.Init(FileName: String);
begin
  inherited Init;
  New(Tokens, Init(FileName));
  New(JadObject, Init);
end; {TJadsCompiler.Init}

destructor TJadsCompiler.Done;
begin
  Dispose(JadObject, Done);
  Dispose(Tokens, Done);
  inherited Done;
end; {TJadsCompiler.Done}

procedure TJadsCompiler.Error(CompilerError: TCompilerError);
begin
  Writeln(
    Tokens^.InputFileName, '(',
    Token.Line, ',',
    Token.Position,
    '): Fatal: ', ErrorMessages[CompilerError]);
  Halt(256-2);
end; {TJadsCompiler.Error}

procedure TJadsCompiler.Compile;
begin
  if Tokens^.Status <> 0 then
    Error(ceFileNotFound);

  CompileObjectBody;
  if not CompileEndOfFile then
    Error(ceEndOfFileExpected);

  Lines := Tokens^.InStream.CurrentLine;
end; {TJadsCompiler.Compile}

function TJadsCompiler.CompileObjectBody: Boolean;
begin
  if not CompileType then;
  if not CompileAttributes then;
  while CompileObject do;
  CompileObjectBody := True;
end; {TJadsCompiler.CompileObjectBody}

function TJadsCompiler.CompileType: Boolean;
var
  Temp: Boolean;
begin
  Temp := False;

  with Tokens^ do
    if ReadReservedWord(Token) then
      if Token.N = rwType then
        if ReadDelimiter(Token) then
          if Delimiters[Token.I] = '=' then
            if ReadString(Token, sStringChars) then
            begin
              {Token.S = "object type"}
              JadObject^.SetType(Token.S);
              Temp := True;
            end
            else
              Error(ceStringConstantExpected)
          else
            Error(ceEqualExpected)
        else
          Error(ceEqualExpected)
      else
        PutBack(Token);

  CompileType := Temp;
end; {TJadsCompiler.CompileType}

function TJadsCompiler.CompileAttributes: Boolean;
var
  Temp: Boolean;
begin
  Temp := False;

  with Tokens^ do
    { "attributes" }
    if ReadReservedWord(Token) then
      if Token.N = rwAttributes then
        { "=" }
        if ReadDelimiter(Token) then
          if Delimiters[Token.I] = '=' then
          begin
            { "(" }
            if ReadDelimiter(Token) then
              if Delimiters[Token.I] = '(' then
              else
                Error(ceLeftParenthesisExpected)
            else
              Error(ceLeftParenthesisExpected);

            { <attribute> }
            while CompileAttribute do;

            { ")" }
            if ReadDelimiter(Token) then
              if Delimiters[Token.I] = ')' then
              else
                Error(ceRightParenthesisExpected)
            else
              Error(ceRightParenthesisExpected);
          end
          else
            Error(ceEqualExpected)
        else
          Error(ceEqualExpected)
      else
        PutBack(Token);

  CompileAttributes := Temp;
end; {TJadsCompiler.CompileAttributes}

function TJadsCompiler.CompileObject: Boolean;
var
  Temp: Boolean;
  TempObject: PJadObject;
begin
  Temp := False;

  with Tokens^ do
    { "object" }
    if ReadReservedWord(Token) then
      if Token.N = rwObject then
        { "=" }
        if ReadDelimiter(Token) then
          if Delimiters[Token.I] = '=' then
          begin
            { "(" }
            if ReadDelimiter(Token) then
              if Delimiters[Token.I] = '(' then
              else
                Error(ceLeftParenthesisExpected)
            else
              Error(ceLeftParenthesisExpected);

            { create new data object and exchange links }
            TempObject := JadObject;
            JadObject := New(PJadObject, Init);

            { <objectbody> }
            CompileObjectBody;
            Temp := True;

            { insert new data object into the root object }
            TempObject^.AddObject(JadObject);
            JadObject := TempObject;

            { ")" }
            if ReadDelimiter(Token) then
              if Delimiters[Token.I] = ')' then
              else
                Error(ceRightParenthesisExpected)
            else
              Error(ceRightParenthesisExpected);
          end
          else
            Error(ceEqualExpected)
        else
          Error(ceEqualExpected)
      else
        PutBack(Token);

  CompileObject := Temp;
end; {TJadsCompiler.CompileObject}

function TJadsCompiler.CompileAttribute: Boolean;
var
  Temp: Boolean;
  Name, Value: String;
begin
  Temp := False;

  with Tokens^ do
    if ReadIdentifier(Token) then
    begin
      Name := Token.S;

      if ReadDelimiter(Token) then
        if Delimiters[Token.I] = '=' then
        else
          Error(ceEqualExpected)
      else
        Error(ceEqualExpected);

      if ReadString(Token, sStringChars) then
        Value := Token.S
      else
        Error(ceStringConstantExpected);

      {got Name=Value pair, so add it to attribute list}
      JadObject^.AddAttribute(Name, Value);
      Temp := True;
    end;

  CompileAttribute := Temp;
end; {TJadsCompiler.CompileAttribute}

function TJadsCompiler.CompileEndOfFile: Boolean;
var
  Temp: Boolean;
begin
  Temp := True;
  Token.Kind := tkUnknown;

  with Tokens^ do
    if Read(Token) then
      Temp := False
    else
      if (InStream.Status = 0) then
        Temp := False;

  CompileEndOfFile := Temp;
end; {TJadsCompiler.CompileEndOfFile}

end {CSyntax}.
